from inewave.newave.modelos.parpeol import (
    BlocoSerieVentosUEE,
    BlocoCorrelVentosUEE,
    BlocoSerieRuidosUEE,
    BlocoCorrelRuidosUEE,
    BlocoCorrelEspacialAnualConfig,
    BlocoCorrelEspacialMensalConfig,
)

from inewave.newave.parpeol import PARpeol


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.parpeol import (
    MockSeriesVentosUEE,
    MockCorrelacaoSeriesVentosUEE,
    MockSerieRuidosUEE,
    MockCorrelSerieRuidosUEE,
    MockCorrelEspacialAnual,
    MockCorrelEspacialMensal,
    MockPARpeol,
)


def test_series_ventos_uee():

    m: MagicMock = mock_open(read_data="".join(MockSeriesVentosUEE))
    b = BlocoSerieVentosUEE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 40
    assert b.data.shape[1] == 15
    assert b.data.iloc[0, 3] == 6.31
    assert b.data.iloc[-1, -1] == 6.93


def test_correlacao_series_ventos_uee():

    m: MagicMock = mock_open(read_data="".join(MockCorrelacaoSeriesVentosUEE))
    b = BlocoCorrelVentosUEE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 12
    assert b.data.iloc[0, 1] == -0.06591
    assert b.data.iloc[-1, -1] == 0.29579


def test_serie_ruidos_uee():

    m: MagicMock = mock_open(read_data="".join(MockSerieRuidosUEE))
    b = BlocoSerieRuidosUEE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 40
    assert b.data.shape[1] == 14
    assert b.data.iloc[0, 1] == 0.0
    assert b.data.iloc[-1, -1] == 0.497


def test_correl_serie_ruidos_uee():

    m: MagicMock = mock_open(read_data="".join(MockCorrelSerieRuidosUEE))
    b = BlocoCorrelRuidosUEE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 120
    assert b.data.shape[1] == 12
    assert b.data.iloc[0, 1] == 0.29579
    assert b.data.iloc[-1, -1] == 0.00582


def test_correl_espacial_anual():

    m: MagicMock = mock_open(read_data="".join(MockCorrelEspacialAnual))
    b = BlocoCorrelEspacialAnualConfig()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 2
    assert b.data.shape[1] == 16
    assert b.data.iloc[0, 2] == -0.2010
    assert b.data.iloc[-1, -1] == 1.000


def test_correl_espacial_mensal():

    m: MagicMock = mock_open(read_data="".join(MockCorrelEspacialMensal))
    b = BlocoCorrelEspacialMensalConfig()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 24
    assert b.data.shape[1] == 17
    assert b.data.iloc[0, 3] == -0.3616
    assert b.data.iloc[-1, -1] == 1.000


def test_atributos_encontrados_parpeol():
    m: MagicMock = mock_open(read_data="".join(MockPARpeol))
    with patch("builtins.open", m):
        parp = PARpeol.le_arquivo("")
        assert parp.series_ventos_uee is not None
        assert parp.correlacao_series_ventos_uee is not None
        assert parp.series_ruido_uee is not None
        assert parp.correlacao_series_ruidos_uee is not None
        assert parp.correlacao_espacial_anual is not None
        assert parp.correlacao_espacial_mensal is not None


def test_atributos_nao_encontrados_parp():
    m: MagicMock = mock_open(read_data="".join(""))
    with patch("builtins.open", m):
        parp = PARpeol.le_arquivo("")
        assert parp.series_ventos_uee is None
        assert parp.correlacao_series_ventos_uee is None
        assert parp.series_ruido_uee is None
        assert parp.correlacao_series_ruidos_uee is None
        assert parp.correlacao_espacial_anual is None
        assert parp.correlacao_espacial_mensal is None


def test_eq_parpeol():
    m: MagicMock = mock_open(read_data="".join(MockPARpeol))
    with patch("builtins.open", m):
        parp1 = PARpeol.le_arquivo("")
        parp2 = PARpeol.le_arquivo("")
        assert parp1 == parp2


def test_neq_parpeol():
    m: MagicMock = mock_open(read_data="".join(MockPARpeol))
    with patch("builtins.open", m):
        parp1 = PARpeol.le_arquivo("")
        parp2 = PARpeol.le_arquivo("")
        parp2.series_ventos_uee.iloc[0, 0] = -1
        assert parp1 != parp2



from inewave.nwlistop.modelos.ctermsin import CtermsAnos
from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class CtermSIN(ArquivoSIN):
    """
    Armazena os dados das saídas referentes aos custos de geração térmica
    para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `ctermsin.out`.

    """

    BLOCKS = [
        CtermsAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="ctermsin.out"
    ) -> "CtermSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="ctermsin.out"):
        self.write(diretorio, nome_arquivo)



from inewave.newave.modelos.newavetim import BlocoTemposEtapasTim

from cfinterface.components.block import Block
from cfinterface.files.blockfile import BlockFile
from typing import Type, TypeVar, Optional, Any
import pandas as pd  # type: ignore


class NewaveTim(BlockFile):
    """
    Armazena os dados de saída do NEWAVE referentes aos
    tempos de execução do programa.

    """

    T = TypeVar("T")

    BLOCKS = [BlocoTemposEtapasTim]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="newave.tim"
    ) -> "NewaveTim":
        return cls.read(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    def __extrai_dados_se_existe(
        self, bloco: Type[Block], indice: int = 0
    ) -> Optional[Any]:
        """
        Obtém os dados de um bloco se este existir dentre os blocos do arquivo.

        :param bloco: O tipo do bloco cujos dados serão extraídos
        :type bloco: Type[T]
        :param indice: Qual dos blocos do tipo será acessado
        :type indice: int, optional
        :return: Os dados do bloco, se existirem
        :rtype: Any
        """
        b = self.__bloco_por_tipo(bloco, indice)
        if b is not None:
            return b.data
        return None

    @property
    def tempos_etapas(self) -> Optional[pd.DataFrame]:
        """
        Tempos de execução do modelo por etapa do modelo.

        - Etapa (`str`)
        - Tempo (`timedelta`)

        :return: Os tempos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        return self.__extrai_dados_se_existe(BlocoTemposEtapasTim)



MockRegistroEolicaSubmercado = "EOLICA-SUBMERCADO ; 5 ; 2\n"

MockEolicaSubmercado = [
    "& EOLICA-SUBMERCADO; CodigoEolica; CodigoSubmercado\n",
    "EOLICA-SUBMERCADO ; 1 ; 3\n",
    "EOLICA-SUBMERCADO ; 2 ; 3\n",
    "EOLICA-SUBMERCADO ; 3 ; 3\n",
    "EOLICA-SUBMERCADO ; 4 ; 2\n",
    "EOLICA-SUBMERCADO ; 5 ; 2\n",
    "",
]



from inewave.newave.modelos.arquivos import BlocoNomesArquivos

from cfinterface.files.sectionfile import SectionFile

from typing import List, TypeVar, Type, Optional


class Arquivos(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao arquivo
    `arquivos.dat`.

    Esta classe lida com informações de entrada do NEWAVE e
    que deve se referir aos nomes dos demais arquivos de entrada
    utilizados para o caso em questão.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoNomesArquivos]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="arquivos.dat"
    ) -> "Arquivos":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="arquivos.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    def __le_nome_por_indice(self, indice: int) -> Optional[str]:
        b = self.__bloco_por_tipo(BlocoNomesArquivos, 0)
        if b is not None:
            if indice in b.data.index:
                dado = b.data.iloc[indice, 1]
                if isinstance(dado, str):
                    return dado
        return None

    def __atualiza_nome_por_indice(self, indice: int, nome: str):
        b = self.__bloco_por_tipo(BlocoNomesArquivos, 0)
        if b is not None:
            b.data.iloc[indice, 1] = nome

    @property
    def arquivos(self) -> List[str]:
        """
        Os nomes dos arquivos utilizados.

        :return: Os arquivos na mesma ordem em que são declarados
        :rtype: List[str]
        """
        b = self.__bloco_por_tipo(BlocoNomesArquivos, 0)
        return [] if b is None else b.data.iloc[:, 1]

    @property
    def arquivos_entrada(self) -> List[str]:
        """
        Os nomes dos arquivos de entrada utilizados.

        :return: Os arquivos de entrada
            na mesma ordem em que são declarados
        :rtype: List[str]
        """
        b = self.__bloco_por_tipo(BlocoNomesArquivos, 0)
        if b is None:
            return []
        todos_indices = set(list(range(len(b.data.index.tolist()))))
        indices_saida = set([10, 11, 12, 13, 14, 15, 18])
        indices_entrada = list(todos_indices.difference(indices_saida))
        return b.data.iloc[indices_entrada, 1]

    @property
    def dger(self) -> Optional[str]:
        """
        Nome do arquivo de dados gerais utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(0)

    @dger.setter
    def dger(self, arq: str):
        self.__atualiza_nome_por_indice(0, arq)

    @property
    def sistema(self) -> Optional[str]:
        """
        Nome do arquivo de subsistemas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(1)

    @sistema.setter
    def sistema(self, arq: str):
        self.__atualiza_nome_por_indice(1, arq)

    @property
    def confhd(self) -> Optional[str]:
        """
        Nome do arquivo de configuração hidráulica utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(2)

    @confhd.setter
    def confhd(self, arq: str):
        self.__atualiza_nome_por_indice(2, arq)

    @property
    def modif(self) -> Optional[str]:
        """
        Nome do arquivo de modificações hidráulicas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(3)

    @modif.setter
    def modif(self, arq: str):
        self.__atualiza_nome_por_indice(3, arq)

    @property
    def conft(self) -> Optional[str]:
        """
        Nome do arquivo de configuração térmica utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(4)

    @conft.setter
    def conft(self, arq: str):
        self.__atualiza_nome_por_indice(4, arq)

    @property
    def term(self) -> Optional[str]:
        """
        Nome do arquivo de dados de térmicas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(5)

    @term.setter
    def term(self, arq: str):
        self.__atualiza_nome_por_indice(5, arq)

    @property
    def clast(self) -> Optional[str]:
        """
        Nome do arquivo de classes térmicas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(6)

    @clast.setter
    def clast(self, arq: str):
        self.__atualiza_nome_por_indice(6, arq)

    @property
    def exph(self) -> Optional[str]:
        """
        Nome do arquivo de expansão hidráulica utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(7)

    @exph.setter
    def exph(self, arq: str):
        self.__atualiza_nome_por_indice(7, arq)

    @property
    def expt(self) -> Optional[str]:
        """
        Nome do arquivo de expansão térmica utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(8)

    @expt.setter
    def expt(self, arq: str):
        self.__atualiza_nome_por_indice(8, arq)

    @property
    def patamar(self) -> Optional[str]:
        """
        Nome do arquivo de patamares de mercado utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(9)

    @patamar.setter
    def patamar(self, arq: str):
        self.__atualiza_nome_por_indice(9, arq)

    @property
    def cortes(self) -> Optional[str]:
        """
        Nome do arquivo de cortes utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(10)

    @cortes.setter
    def cortes(self, arq: str):
        self.__atualiza_nome_por_indice(10, arq)

    @property
    def cortesh(self) -> Optional[str]:
        """
        Nome do arquivo de cabeçalho de cortes utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(11)

    @cortesh.setter
    def cortesh(self, arq: str):
        self.__atualiza_nome_por_indice(11, arq)

    @property
    def pmo(self) -> Optional[str]:
        """
        Nome do arquivo de relatório de execução utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(12)

    @pmo.setter
    def pmo(self, arq: str):
        self.__atualiza_nome_por_indice(12, arq)

    @property
    def parp(self) -> Optional[str]:
        """
        Nome do arquivo de séries sintéticas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(13)

    @parp.setter
    def parp(self, arq: str):
        self.__atualiza_nome_por_indice(13, arq)

    @property
    def forward(self) -> Optional[str]:
        """
        Nome do arquivo de relatório da forward utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(14)

    @forward.setter
    def forward(self, arq: str):
        self.__atualiza_nome_por_indice(14, arq)

    @property
    def forwardh(self) -> Optional[str]:
        """
        Nome do arquivo de cabeçalho da forward utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(15)

    @forwardh.setter
    def forwardh(self, arq: str):
        self.__atualiza_nome_por_indice(15, arq)

    @property
    def shist(self) -> Optional[str]:
        """
        Nome do arquivo de séries históricas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(16)

    @shist.setter
    def shist(self, arq: str):
        self.__atualiza_nome_por_indice(16, arq)

    @property
    def manutt(self) -> Optional[str]:
        """
        Nome do arquivo de programação da manutenção de
        térmicas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(17)

    @manutt.setter
    def manutt(self, arq: str):
        self.__atualiza_nome_por_indice(17, arq)

    @property
    def newdesp(self) -> Optional[str]:
        """
        Nome do arquivo de despacho hidrotérmico utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(18)

    @newdesp.setter
    def newdesp(self, arq: str):
        self.__atualiza_nome_por_indice(18, arq)

    @property
    def vazpast(self) -> Optional[str]:
        """
        Nome do arquivo de tendência hidrológica utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(19)

    @vazpast.setter
    def vazpast(self, arq: str):
        self.__atualiza_nome_por_indice(19, arq)

    @property
    def itaipu(self) -> Optional[str]:
        """
        Nome do arquivo de dados de Itaipu utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(20)

    @itaipu.setter
    def itaipu(self, arq: str):
        self.__atualiza_nome_por_indice(20, arq)

    @property
    def bid(self) -> Optional[str]:
        """
        Nome do arquivo de bidding utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(21)

    @bid.setter
    def bid(self, arq: str):
        self.__atualiza_nome_por_indice(21, arq)

    @property
    def c_adic(self) -> Optional[str]:
        """
        Nome do arquivo de cargas adicionais utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(22)

    @c_adic.setter
    def c_adic(self, arq: str):
        self.__atualiza_nome_por_indice(22, arq)

    @property
    def perda(self) -> Optional[str]:
        """
        Nome do arquivo de fatores de perdas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(23)

    @perda.setter
    def perda(self, arq: str):
        self.__atualiza_nome_por_indice(23, arq)

    @property
    def gtminpat(self) -> Optional[str]:
        """
        Nome do arquivo de patamares de geração térmica
        mínima utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(24)

    @gtminpat.setter
    def gtminpat(self, arq: str):
        self.__atualiza_nome_por_indice(24, arq)

    @property
    def elnino(self) -> Optional[str]:
        """
        Nome do arquivo de ENSO 1 utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(25)

    @elnino.setter
    def elnino(self, arq: str):
        self.__atualiza_nome_por_indice(25, arq)

    @property
    def ensoaux(self) -> Optional[str]:
        """
        Nome do arquivo de ENSO 2 utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(26)

    @ensoaux.setter
    def ensoaux(self, arq: str):
        self.__atualiza_nome_por_indice(26, arq)

    @property
    def dsvagua(self) -> Optional[str]:
        """
        Nome do arquivo de desvio de água utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(27)

    @dsvagua.setter
    def dsvagua(self, arq: str):
        self.__atualiza_nome_por_indice(27, arq)

    @property
    def penalid(self) -> Optional[str]:
        """
        Nome do arquivo de penalidades por desvio utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(28)

    @penalid.setter
    def penalid(self, arq: str):
        self.__atualiza_nome_por_indice(28, arq)

    @property
    def curva(self) -> Optional[str]:
        """
        Nome do arquivo com a curva guia de penalidades por
        volume mínimo armazenado utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(29)

    @curva.setter
    def curva(self, arq: str):
        self.__atualiza_nome_por_indice(29, arq)

    @property
    def agrint(self) -> Optional[str]:
        """
        Nome do arquivo de agrupamento livre de intercâmbios
        utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(30)

    @agrint.setter
    def agrint(self, arq: str):
        self.__atualiza_nome_por_indice(30, arq)

    @property
    def adterm(self) -> Optional[str]:
        """
        Nome do arquivo de atencipação de despacho GNL
        utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(31)

    @adterm.setter
    def adterm(self, arq: str):
        self.__atualiza_nome_por_indice(31, arq)

    @property
    def ghmin(self) -> Optional[str]:
        """
        Nome do arquivo de geração hidraulica mínima utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(32)

    @ghmin.setter
    def ghmin(self, arq: str):
        self.__atualiza_nome_por_indice(32, arq)

    @property
    def sar(self) -> Optional[str]:
        """
        Nome do arquivo de aversão a risco SAR utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(33)

    @sar.setter
    def sar(self, arq: str):
        self.__atualiza_nome_por_indice(33, arq)

    @property
    def cvar(self) -> Optional[str]:
        """
        Nome do arquivo de aversão a risco CVAR utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(34)

    @cvar.setter
    def cvar(self, arq: str):
        self.__atualiza_nome_por_indice(34, arq)

    @property
    def ree(self) -> Optional[str]:
        """
        Nome do arquivo de configuração das REEs utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(35)

    @ree.setter
    def ree(self, arq: str):
        self.__atualiza_nome_por_indice(35, arq)

    @property
    def re(self) -> Optional[str]:
        """
        Nome do arquivo de restrições elétricas utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(36)

    @re.setter
    def re(self, arq: str):
        self.__atualiza_nome_por_indice(36, arq)

    @property
    def tecno(self) -> Optional[str]:
        """
        Nome do arquivo de tecnologias utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(37)

    @tecno.setter
    def tecno(self, arq: str):
        self.__atualiza_nome_por_indice(37, arq)

    @property
    def abertura(self) -> Optional[str]:
        """
        Nome do arquivo de aberturas por período utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(38)

    @abertura.setter
    def abertura(self, arq: str):
        self.__atualiza_nome_por_indice(38, arq)

    @property
    def gee(self) -> Optional[str]:
        """
        Nome do arquivo de emissões GEE utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(39)

    @gee.setter
    def gee(self, arq: str):
        self.__atualiza_nome_por_indice(39, arq)

    @property
    def clasgas(self) -> Optional[str]:
        """
        Nome do arquivo de restrições de gás utilizado pelo NEWAVE.
        """
        return self.__le_nome_por_indice(40)

    @clasgas.setter
    def clasgas(self, arq: str):
        self.__atualiza_nome_por_indice(40, arq)



# Rotinas de testes associadas ao arquivo re.dat do NEWAVE
from inewave.newave.modelos.re import (
    BlocoUsinasConjuntoRE,
    BlocoConfiguracaoRestricoesRE,
)

from inewave.newave import RE


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.re import (
    MockBlocoUsinasRestricoes,
    MockBlocoRestricoes,
    MockRE,
)


def test_bloco_usinas_conjuntos_re():

    m: MagicMock = mock_open(read_data="".join(MockBlocoUsinasRestricoes))
    b = BlocoUsinasConjuntoRE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, 0] == 10
    assert b.data.iloc[-1, 2] == 284


def test_bloco_restricoes_re():

    m: MagicMock = mock_open(read_data="".join(MockBlocoRestricoes))
    b = BlocoConfiguracaoRestricoesRE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, 0] == 10
    assert b.data.iloc[-1, -1] == "C. CALDEIRAO E F. GOMES"


def test_atributos_encontrados_re():
    m: MagicMock = mock_open(read_data="".join(MockRE))
    with patch("builtins.open", m):
        ad = RE.le_arquivo("")
        assert ad.usinas_conjuntos is not None
        assert ad.restricoes is not None


def test_atributos_nao_encontrados_re():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = RE.le_arquivo("")
        assert ad.usinas_conjuntos is None
        assert ad.restricoes is None


def test_eq_re():
    m: MagicMock = mock_open(read_data="".join(MockRE))
    with patch("builtins.open", m):
        cf1 = RE.le_arquivo("")
        cf2 = RE.le_arquivo("")
        assert cf1 == cf2


def test_neq_re():
    m: MagicMock = mock_open(read_data="".join(MockRE))
    with patch("builtins.open", m):
        cf1 = RE.le_arquivo("")
        cf2 = RE.le_arquivo("")
        cf2.usinas_conjuntos.loc[0, 0] = 0
        assert cf1 != cf2


def test_leitura_escrita_re():
    m_leitura: MagicMock = mock_open(read_data="".join(MockRE))
    with patch("builtins.open", m_leitura):
        cf1 = RE.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = RE.le_arquivo("")
        assert cf1 == cf2



from inewave.newave.modelos.caso import NomeCaso, CaminhoGerenciadorProcessos
from inewave.newave import Caso

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.caso import (
    MockNomeCaso,
    MockCaminhoGerenciador,
    MockCaso,
)


def test_bloco_nome_caso():

    m: MagicMock = mock_open(read_data="".join(MockNomeCaso))
    b = NomeCaso()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == "arquivos.dat"


def test_bloco_caminho_gerenciador_caso():

    m: MagicMock = mock_open(read_data="".join(MockCaminhoGerenciador))
    b = CaminhoGerenciadorProcessos()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == "/home/newave/"


def test_atributos_encontrados_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad = Caso.le_arquivo("")
        assert ad.arquivos is not None
        assert ad.gerenciador_processos is not None


def test_atributos_nao_encontrados_caso():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Caso.le_arquivo("")
        assert ad.arquivos == ""
        assert ad.gerenciador_processos == ""


def test_eq_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad1 = Caso.le_arquivo("")
        ad2 = Caso.le_arquivo("")
        assert ad1 == ad2


def test_neq_caso():
    m: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m):
        ad1 = Caso.le_arquivo("")
        ad2 = Caso.le_arquivo("")
        ad2.gerenciador_processos = "/bin"
        assert ad1 != ad2


def test_leitura_escrita_caso():
    m_leitura: MagicMock = mock_open(read_data="".join(MockCaso))
    with patch("builtins.open", m_leitura):
        ad1 = Caso.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Caso.le_arquivo("")
        assert ad1 == ad2



from inewave.nwlistop.modelos.coper import CoperAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class Coper(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao custo total de operação
    de cada estágio em cada série.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `coper.out`.
    """

    BLOCKS = [
        CoperAnos,
    ]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="coper.out") -> "Coper":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="coper.out"):
        self.write(diretorio, nome_arquivo)



# Rotinas de testes associadas ao arquivo cvar.dat do NEWAVE
from inewave.newave.modelos.cvar import (
    BlocoValoresConstantesCVAR,
    BlocoAlfaVariavelNoTempo,
    BlocoLambdaVariavelNoTempo,
)

from inewave.newave import CVAR


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.cvar import (
    MockBlocoValoresConstantes,
    MockBlocoValoresAlfaVariaveis,
    MockBlocoValoresLambdaVariaveis,
    MockCVAR,
)


def test_bloco_valores_constantes_cvar():

    m: MagicMock = mock_open(read_data="".join(MockBlocoValoresConstantes))
    b = BlocoValoresConstantesCVAR()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == [50.0, 40.0]


def test_bloco_alfa_variavel_cvar():

    m: MagicMock = mock_open(read_data="".join(MockBlocoValoresAlfaVariaveis))
    b = BlocoAlfaVariavelNoTempo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == "2017"
    assert b.data.iloc[-1, -1] == 5.0


def test_bloco_lambda_variavel_cvar():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoValoresLambdaVariaveis)
    )
    b = BlocoLambdaVariavelNoTempo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == "2017"
    assert b.data.iloc[-1, -1] == 10.0


def test_atributos_encontrados_cvar():
    m: MagicMock = mock_open(read_data="".join(MockCVAR))
    with patch("builtins.open", m):
        ad = CVAR.le_arquivo("")
        assert ad.valores_constantes != [None, None]
        assert ad.alfa_variavel is not None
        assert ad.lambda_variavel is not None


def test_atributos_nao_encontrados_cvar():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = CVAR.le_arquivo("")
        assert ad.valores_constantes is None
        assert ad.alfa_variavel is None
        assert ad.lambda_variavel is None


def test_eq_cvar():
    m: MagicMock = mock_open(read_data="".join(MockCVAR))
    with patch("builtins.open", m):
        cf1 = CVAR.le_arquivo("")
        cf2 = CVAR.le_arquivo("")
        assert cf1 == cf2


def test_neq_cvar():
    m: MagicMock = mock_open(read_data="".join(MockCVAR))
    with patch("builtins.open", m):
        cf1 = CVAR.le_arquivo("")
        cf2 = CVAR.le_arquivo("")
        cf2.valores_constantes = [0, 0]
        assert cf1 != cf2


def test_leitura_escrita_cvar():
    m_leitura: MagicMock = mock_open(read_data="".join(MockCVAR))
    with patch("builtins.open", m_leitura):
        cf1 = CVAR.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = CVAR.le_arquivo("")
        assert cf1 == cf2



from inewave.newave.modelos.dsvagua import BlocoDsvUHE

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class DSVAgua(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    desvios de água por usina.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `dsvagua.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoDsvUHE]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="dsvagua.dat"
    ) -> "DSVAgua":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="dsvagua.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def desvios(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os desvios de água por usina e por estágio.

        - Ano (`int`)
        - Usina (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)
        - Flag (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoDsvUHE, 0)
        if b is not None:
            return b.data
        return None

    @desvios.setter
    def desvios(self, valor: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoDsvUHE, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from inewave.newave.modelos.patamar import (
    BlocoNumeroPatamares,
    BlocoDuracaoPatamar,
    BlocoCargaPatamar,
    BlocoIntercambioPatamarSubsistemas,
    BlocoUsinasNaoSimuladas,
)

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class Patamar(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    patamares de carga por submercado.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoNumeroPatamares,
        BlocoDuracaoPatamar,
        BlocoCargaPatamar,
        BlocoIntercambioPatamarSubsistemas,
        BlocoUsinasNaoSimuladas,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="patamar.dat"
    ) -> "Patamar":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="patamar.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def numero_patamares(self) -> Optional[int]:
        """
        O número de patamares utilizado no estudo.

        :return: O número de patamares como um inteiro
        :rtype: int | None
        """
        b = self.__bloco_por_tipo(BlocoNumeroPatamares, 0)
        if b is not None:
            return b.data
        return None

    @numero_patamares.setter
    def numero_patamares(self, n: int):
        b = self.__bloco_por_tipo(BlocoNumeroPatamares, 0)
        if b is not None:
            b.data = n

    @property
    def duracao_mensal_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a duração mensal de cada patamar no horizonte
        de estudo.

        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoDuracaoPatamar, 0)
        if b is not None:
            return b.data
        return None

    @duracao_mensal_patamares.setter
    def duracao_mensal_patamares(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoDuracaoPatamar, 0)
        if b is not None:
            b.data = df

    @property
    def carga_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a carga em P.U. por patamar.

        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoCargaPatamar, 0)
        if b is not None:
            return b.data
        return None

    @carga_patamares.setter
    def carga_patamares(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoCargaPatamar, 0)
        if b is not None:
            b.data = df

    @property
    def intercambio_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a correção em P.U. do intercâmbio por patamar.

        - Subsistema De (`str`)
        - Subsistema Para (`str`)
        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoIntercambioPatamarSubsistemas, 0)
        if b is not None:
            return b.data
        return None

    @intercambio_patamares.setter
    def intercambio_patamares(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoIntercambioPatamarSubsistemas, 0)
        if b is not None:
            b.data = df

    @property
    def usinas_nao_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os fatores das usinas não simuladas em P.U.

        - Subsistema (`str`)
        - Bloco (`int`)
        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: Os valores por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoUsinasNaoSimuladas, 0)
        if b is not None:
            return b.data
        return None

    @usinas_nao_simuladas.setter
    def usinas_nao_simuladas(self, df: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoUsinasNaoSimuladas, 0)
        if b is not None:
            b.data = df



from inewave.config import MAX_ANOS_ESTUDO, MESES_DF

from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.floatfield import FloatField
from cfinterface.components.literalfield import LiteralField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoValoresConstantesCVAR(Block):
    """
    Bloco com valores dos parâmetros ALFA e LAMBDA constantes.
    """

    BEGIN_PATTERN = "VALORES CONSTANTE NO TEMPO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                FloatField(5, 7, 1),
                FloatField(5, 14, 1),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoValoresConstantesCVAR):
            return False
        bloco: BlocoValoresConstantesCVAR = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        for _ in range(2):
            self.__cabecalhos.append(file.readline())
        self.data = self.__linha.read(file.readline())

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, list):
            raise ValueError("Dados do cvar.dat não foram lidos com sucesso")
        file.write(self.__linha.write(self.data))


class BlocoAlfaVariavelNoTempo(Block):
    """
    Bloco com a informação do valor de ALFA por estágio
    no horizonte de execução.
    """

    BEGIN_PATTERN = "VALORES DE ALFA VARIAVEIS NO TEMPO"
    END_PATTERN = "VALORES DE LAMBDA VARIAVEIS NO TEMPO"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [LiteralField(5, 0)]
        campos_valores: List[Field] = [
            FloatField(5, 7 * i + 7, 1) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_valores)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoAlfaVariavelNoTempo):
            return False
        bloco: BlocoAlfaVariavelNoTempo = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["Ano"] = anos
            df = df
            df = df[["Ano"] + MESES_DF]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO, len(MESES_DF)))
        anos: List[str] = []
        while True:
            ultima_linha = file.tell()
            linha = file.readline()
            if len(linha) < 3:
                break
            # Confere se terminaram
            if self.ends(linha):
                file.seek(ultima_linha)
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do cvar.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha.write(lin.tolist()))


class BlocoLambdaVariavelNoTempo(Block):
    """
    Bloco com a informação do valor de LAMBDA por estágio
    no horizonte de execução.
    """

    BEGIN_PATTERN = "VALORES DE LAMBDA VARIAVEIS NO TEMPO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [LiteralField(5, 0)]
        campos_valores: List[Field] = [
            FloatField(5, 7 * i + 7, 1) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_valores)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoLambdaVariavelNoTempo):
            return False
        bloco: BlocoLambdaVariavelNoTempo = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["Ano"] = anos
            df = df
            df = df[["Ano"] + MESES_DF]
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO, len(MESES_DF)))
        anos: List[str] = []
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3:
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do cvar.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha.write(lin.tolist()))



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore

from inewave.config import MESES_DF, MAX_UTES


class BlocoTermUTE(Section):
    """
    Bloco de informações das classes de usinas térmicas
    existentes no arquivo do NEWAVE `term.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                LiteralField(12, 5),
                FloatField(5, 19, 0),
                FloatField(4, 25, 0),
                FloatField(6, 31, 2),
                FloatField(6, 38, 2),
            ]
            + [FloatField(6, 45 + 7 * i, 2) for i in range(len(MESES_DF) + 1)]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoTermUTE):
            return False
        bloco: BlocoTermUTE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = (
                [
                    "Potência Instalada",
                    "FC Máximo",
                    "TEIF",
                    "Indisponibilidade Programada",
                ]
                + [f"GT Min {m}" for m in MESES_DF]
                + ["GT Min D+ Anos"]
            )
            df = pd.DataFrame(
                tabela,
                columns=cols,
            )
            df["Número"] = numeros
            df["Nome"] = nomes
            df = df[["Número", "Nome"] + cols]
            df = df.astype({"Número": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_UTES, len(MESES_DF) + 5))
        numeros: List[int] = []
        nomes: List[str] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            tabela[i, :] = dados[2:]
            numeros.append(dados[0])
            nomes.append(dados[1])
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do term.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha.write(lin.tolist()))



from inewave.config import MAX_CORTES, MAX_REES

from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class CortesPeriodoNwlistcf(Block):
    """
    Bloco do arquivo nwlistcf.rel que armazena os cortes de um período.
    """

    BEGIN_PATTERN = "PERIODO: "
    END_PATTERN = "PERIODO: "

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_periodo = Line([IntegerField(4, 19)])
        campos_iniciais: List[Field] = [
            IntegerField(8, 2),
            IntegerField(3, 11),
            FloatField(16, 15, 4),
        ]
        campos_pis: List[Field] = [
            FloatField(17, 31 + 18 * i, 9) for i in range(15)
        ]
        self.__linha = Line(campos_iniciais + campos_pis)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, CortesPeriodoNwlistcf):
            return False
        bloco: CortesPeriodoNwlistcf = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            cols = (
                ["IREG", "REE", "RHS", "PIV"]
                + [f"PIH({i})" for i in range(1, 7)]
                + [
                    f"PIGTAD(P{i}L{j})"
                    for i in range(1, 4)
                    for j in range(1, 3)
                ]
                + ["PIMX_SAR", "PIMX_VMN"]
            )
            df = pd.DataFrame(tabela, columns=cols)
            df = df.astype({"IREG": "int64", "REE": "int64"})
            df["PERIODO"] = self.__periodo
            df = df[["PERIODO"] + cols]
            return df

        # Lê o período e as linhas de cabeçalho
        self.__periodo = self.__linha_periodo.read(file.readline())[0]
        for _ in range(2):
            file.readline()

        # Lê as linhas de cortes
        self.__ireg_atual = 0
        tabela = np.zeros((MAX_CORTES * MAX_REES, 18))
        i = 0
        while True:
            ultima_posicao = file.tell()
            linha = file.readline()
            if self.ends(linha) or len(linha) < 3:
                file.seek(ultima_posicao)
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            if dados[0] is not None:
                self.__ireg_atual = dados[0]
            tabela[i, 0] = self.__ireg_atual
            tabela[i, 1:] = dados[1:]
            i += 1



from inewave.nwlistop.modelos.earmfpsin import EarmsAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class EarmfpSIN(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às energias
    armazenadas finais para o SIN e em % da energia armazenável máxima.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `earmfpsin.out`, onde x varia conforme o
    submercado em questão.
    """

    BLOCKS = [
        EarmsAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="earmfpsin.out"
    ) -> "EarmfpSIN":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="earmfpsin.out"):
        self.write(diretorio, nome_arquivo)



# Rotinas de testes associadas ao arquivo sistema.dat do NEWAVE
from inewave.newave.modelos.sistema import (
    BlocoNumeroPatamaresDeficit,
    BlocoCustosDeficit,
    BlocoIntercambioSubsistema,
    BlocoMercadoEnergiaSistema,
    BlocoGeracaoUsinasNaoSimuladas,
)

from inewave.newave import Sistema


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.sistema import (
    MockBlocoNumeroPatamaresDeficit,
    MockBlocoCustoDeficit,
    MockBlocoLimitesIntercambio,
    MockBlocoMercadoEnergia,
    MockBlocoGeracaoUsinasNaoSimuladas,
    MockSistema,
)


def test_bloco_numero_patamares_deficit_sistema():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoNumeroPatamaresDeficit)
    )
    b = BlocoNumeroPatamaresDeficit()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 1


def test_bloco_custos_deficit_sistema():

    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoDeficit))
    b = BlocoCustosDeficit()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, 0] == 11
    assert b.data.iloc[0, 3] == 6524.05


def test_bloco_limites_intercambio_sistema():

    m: MagicMock = mock_open(read_data="".join(MockBlocoLimitesIntercambio))
    b = BlocoIntercambioSubsistema()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 7500.0


def test_bloco_mercado_energia_sistema():

    m: MagicMock = mock_open(read_data="".join(MockBlocoMercadoEnergia))
    b = BlocoMercadoEnergiaSistema()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 6657.0


def test_bloco_usinas_nao_simuladas_sistema():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoGeracaoUsinasNaoSimuladas)
    )
    b = BlocoGeracaoUsinasNaoSimuladas()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 1.0


def test_atributos_encontrados_sistema():
    m: MagicMock = mock_open(read_data="".join(MockSistema))
    with patch("builtins.open", m):
        ad = Sistema.le_arquivo("")
        assert ad.numero_patamares_deficit is not None
        assert ad.custo_deficit is not None
        assert ad.limites_intercambio is not None
        assert ad.mercado_energia is not None
        assert ad.geracao_usinas_nao_simuladas is not None


def test_atributos_nao_encontrados_sistema():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Sistema.le_arquivo("")
        assert ad.numero_patamares_deficit is None
        assert ad.custo_deficit is None
        assert ad.limites_intercambio is None
        assert ad.mercado_energia is None
        assert ad.geracao_usinas_nao_simuladas is None


def test_eq_sistema():
    m: MagicMock = mock_open(read_data="".join(MockSistema))
    with patch("builtins.open", m):
        cf1 = Sistema.le_arquivo("")
        cf2 = Sistema.le_arquivo("")
        assert cf1 == cf2


def test_neq_sistema():
    m: MagicMock = mock_open(read_data="".join(MockSistema))
    with patch("builtins.open", m):
        cf1 = Sistema.le_arquivo("")
        cf2 = Sistema.le_arquivo("")
        cf2.numero_patamares_deficit = 0
        assert cf1 != cf2


def test_leitura_escrita_sistema():
    m_leitura: MagicMock = mock_open(read_data="".join(MockSistema))
    with patch("builtins.open", m_leitura):
        cf1 = Sistema.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Sistema.le_arquivo("")
        assert cf1 == cf2



from inewave.config import MESES_DF

from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField


from inewave.nwlistop.modelos.blocos.valoresseriepatamar import (
    ValoresSeriePatamar,
)


class GHAnos(ValoresSeriePatamar):
    """
    Bloco com as informações das tabelas de geração hidráulica.
    """

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [
            IntegerField(4, 1),
            LiteralField(5, 6),
        ]  # type: ignore
        + [
            FloatField(8, 12 + 9 * i, 1) for i in range(len(MESES_DF) + 1)  # type: ignore
        ]
    )



from inewave.nwlistcf import Estados

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.estados import MockEstados


def test_atributos_encontrados_estados():
    m: MagicMock = mock_open(read_data="".join(MockEstados))
    with patch("builtins.open", m):
        n = Estados.le_arquivo("")
        assert n.estados is not None


def test_atributos_nao_encontrados_estados():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Estados.le_arquivo("")
        assert n.estados is None


def test_eq_estados():
    m: MagicMock = mock_open(read_data="".join(MockEstados))
    with patch("builtins.open", m):
        n1 = Estados.le_arquivo("")
        n2 = Estados.le_arquivo("")
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



from inewave.newave.modelos.curva import (
    BlocoConfiguracoesPenalizacaoCurva,
    BlocoPenalidadesViolacaoREECurva,
    BlocoCurvaSegurancaSubsistema,
    BlocoMaximoIteracoesProcessoIterativoEtapa2,
    BlocoIteracaoAPartirProcessoIterativoEtapa2,
    BlocoToleranciaProcessoIterativoEtapa2,
    BlocoImpressaoRelatorioProcessoIterativoEtapa2,
)

from cfinterface.files.sectionfile import SectionFile
from typing import Type, TypeVar, Optional
import pandas as pd  # type: ignore


class Curva(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à curva para
    penalização por volume mínimo dos reservatórios.
    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoConfiguracoesPenalizacaoCurva,
        BlocoPenalidadesViolacaoREECurva,
        BlocoCurvaSegurancaSubsistema,
        BlocoMaximoIteracoesProcessoIterativoEtapa2,
        BlocoIteracaoAPartirProcessoIterativoEtapa2,
        BlocoToleranciaProcessoIterativoEtapa2,
        BlocoImpressaoRelatorioProcessoIterativoEtapa2,
    ]

    def __init__(self, data=...) -> None:
        super().__init__(data)

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="curva.dat") -> "Curva":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="curva.dat"):
        self.write(diretorio, nome_arquivo)

    def __bloco_por_tipo(self, bloco: Type[T], indice: int) -> Optional[T]:
        """
        Obtém um gerador de blocos de um tipo, se houver algum no arquivo.

        :param bloco: Um tipo de bloco para ser lido
        :type bloco: T
        :param indice: O índice do bloco a ser acessado, dentre os do tipo
        :type indice: int
        :return: O gerador de blocos, se houver
        :rtype: Optional[Generator[T], None, None]
        """
        try:
            return next(
                b
                for i, b in enumerate(self.data.of_type(bloco))
                if i == indice
            )
        except StopIteration:
            return None

    @property
    def configuracoes_penalizacao(self) -> Optional[list]:
        """
        Linha de configuração das opções de penalização do
        arquivo curva.dat.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: Optional[list]
        """
        b = self.__bloco_por_tipo(BlocoConfiguracoesPenalizacaoCurva, 0)
        if b is not None:
            return b.data
        return None

    @configuracoes_penalizacao.setter
    def configuracoes_penalizacao(self, valor: list):
        b = self.__bloco_por_tipo(BlocoConfiguracoesPenalizacaoCurva, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def custos_penalidades(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os custos para penalização em cada REE.

        - Subsistema (`int`)
        - Custo (`float`)

        :return: Os custos por REE em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoPenalidadesViolacaoREECurva, 0)
        if b is not None:
            return b.data
        return None

    @custos_penalidades.setter
    def custos_penalidades(self, valor: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoPenalidadesViolacaoREECurva, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def curva_seguranca(self) -> Optional[pd.DataFrame]:
        """
        Tabela da curva de segurança por REE.

        - REE (`int`)
        - Ano (`int`)
        - Janeiro (`float`)
        - Fevereiro (`float`)
        - ...
        - Dezembro (`float`)

        :return: Os valores dos campos da linha como uma lista.
        :rtype: pd.DataFrame | None
        """
        b = self.__bloco_por_tipo(BlocoCurvaSegurancaSubsistema, 0)
        if b is not None:
            return b.data
        return None

    @curva_seguranca.setter
    def curva_seguranca(self, valor: pd.DataFrame):
        b = self.__bloco_por_tipo(BlocoCurvaSegurancaSubsistema, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def maximo_iteracoes_etapa2(self) -> Optional[int]:
        """
        Número máximo de iterações no processo iterativo de 2ª
        etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.__bloco_por_tipo(
            BlocoMaximoIteracoesProcessoIterativoEtapa2, 0
        )
        if b is not None:
            return b.data
        return None

    @maximo_iteracoes_etapa2.setter
    def maximo_iteracoes_etapa2(self, valor: int):
        b = self.__bloco_por_tipo(
            BlocoMaximoIteracoesProcessoIterativoEtapa2, 0
        )
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def iteracao_a_partir_etapa2(self) -> Optional[int]:
        """
        Iteração a partir da qual ocorre o processo iterativo de 2ª
        etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.__bloco_por_tipo(
            BlocoIteracaoAPartirProcessoIterativoEtapa2, 0
        )
        if b is not None:
            return b.data
        return None

    @iteracao_a_partir_etapa2.setter
    def iteracao_a_partir_etapa2(self, valor: int):
        b = self.__bloco_por_tipo(
            BlocoIteracaoAPartirProcessoIterativoEtapa2, 0
        )
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def tolerancia_processo_etapa2(self) -> Optional[float]:
        """
        Tolerância para a execução do processo iterativo de etapa 2.

        :return: O valor como um float
        :rtype: Optional[float]
        """
        b = self.__bloco_por_tipo(BlocoToleranciaProcessoIterativoEtapa2, 0)
        if b is not None:
            return b.data
        return None

    @tolerancia_processo_etapa2.setter
    def tolerancia_processo_etapa2(self, valor: int):
        b = self.__bloco_por_tipo(BlocoToleranciaProcessoIterativoEtapa2, 0)
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")

    @property
    def impressao_relatorio_etapa2(self) -> Optional[int]:
        """
        Opção de impressão ou não do relatório do
        processo iterativo de 2ª etapa.

        :return: O valor como um int
        :rtype: Optional[int]
        """
        b = self.__bloco_por_tipo(
            BlocoImpressaoRelatorioProcessoIterativoEtapa2, 0
        )
        if b is not None:
            return b.data
        return None

    @impressao_relatorio_etapa2.setter
    def impressao_relatorio_etapa2(self, valor: int):
        b = self.__bloco_por_tipo(
            BlocoImpressaoRelatorioProcessoIterativoEtapa2, 0
        )
        if b is not None:
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.cmarg import CmargsAnos
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)


class Cmarg(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes aos custos marginais de operação
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cmarg00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        CmargsAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="cmarg001.out"
    ) -> "Cmarg":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="cmarg001.out"):
        self.write(diretorio, nome_arquivo)



# Rotinas de testes associadas ao arquivo curva.dat do NEWAVE
from inewave.newave.modelos.curva import (
    BlocoConfiguracoesPenalizacaoCurva,
    BlocoPenalidadesViolacaoREECurva,
    BlocoCurvaSegurancaSubsistema,
    BlocoMaximoIteracoesProcessoIterativoEtapa2,
    BlocoIteracaoAPartirProcessoIterativoEtapa2,
    BlocoToleranciaProcessoIterativoEtapa2,
    BlocoImpressaoRelatorioProcessoIterativoEtapa2,
)

from inewave.newave import Curva


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.curva import (
    MockBlockTipoPenalizacao,
    MockBlocoCustoPorSistema,
    MockBlocoCurvaSeguranca,
    MockMaximoIteracoesProcessoIterativoEtapa2,
    MockIteracaoAPartirProcessoIterativoEtapa2,
    MockToleranciaProcessoIterativoEtapa2,
    MockImpressaoRelatorioProcessoIterativoEtapa2,
    MockCurva,
)


def test_bloco_configuracoes_penalizacao_curva():

    m: MagicMock = mock_open(read_data="".join(MockBlockTipoPenalizacao))
    b = BlocoConfiguracoesPenalizacaoCurva()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == [1, 11, 1]


def test_bloco_penalidades_ree():

    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoPorSistema))
    b = BlocoPenalidadesViolacaoREECurva()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 1745.08


def test_bloco_curva_seguranca():

    m: MagicMock = mock_open(read_data="".join(MockBlocoCurvaSeguranca))
    b = BlocoCurvaSegurancaSubsistema()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 10.0


def test_bloco_maximo_iteracoes():

    m: MagicMock = mock_open(
        read_data="".join(MockMaximoIteracoesProcessoIterativoEtapa2)
    )
    b = BlocoMaximoIteracoesProcessoIterativoEtapa2()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 0


def test_bloco_iteracao_a_partir():

    m: MagicMock = mock_open(
        read_data="".join(MockIteracaoAPartirProcessoIterativoEtapa2)
    )
    b = BlocoIteracaoAPartirProcessoIterativoEtapa2()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 10


def test_bloco_tolerancia_processo():

    m: MagicMock = mock_open(
        read_data="".join(MockToleranciaProcessoIterativoEtapa2)
    )
    b = BlocoToleranciaProcessoIterativoEtapa2()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 0.01


def test_bloco_impressao_relatorio():

    m: MagicMock = mock_open(
        read_data="".join(MockImpressaoRelatorioProcessoIterativoEtapa2)
    )
    b = BlocoImpressaoRelatorioProcessoIterativoEtapa2()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == 0


def test_atributos_encontrados_curva():
    m: MagicMock = mock_open(read_data="".join(MockCurva))
    with patch("builtins.open", m):
        ad = Curva.le_arquivo("")
        assert ad.configuracoes_penalizacao != [None, None, None]
        assert ad.custos_penalidades is not None
        assert ad.curva_seguranca is not None
        assert ad.maximo_iteracoes_etapa2 is not None
        assert ad.iteracao_a_partir_etapa2 is not None
        assert ad.tolerancia_processo_etapa2 is not None
        assert ad.impressao_relatorio_etapa2 is not None


def test_atributos_nao_encontrados_curva():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Curva.le_arquivo("")
        assert ad.configuracoes_penalizacao == [None, None, None]
        assert ad.custos_penalidades is None
        assert ad.curva_seguranca is None
        assert ad.maximo_iteracoes_etapa2 is None
        assert ad.iteracao_a_partir_etapa2 is None
        assert ad.tolerancia_processo_etapa2 is None
        assert ad.impressao_relatorio_etapa2 is None


def test_eq_curva():
    m: MagicMock = mock_open(read_data="".join(MockCurva))
    with patch("builtins.open", m):
        cf1 = Curva.le_arquivo("")
        cf2 = Curva.le_arquivo("")
        assert cf1 == cf2


def test_neq_curva():
    m: MagicMock = mock_open(read_data="".join(MockCurva))
    with patch("builtins.open", m):
        cf1 = Curva.le_arquivo("")
        cf2 = Curva.le_arquivo("")
        cf2.curva_seguranca.iloc[0, 0] = -1
        assert cf1 != cf2


def test_leitura_escrita_curva():
    m_leitura: MagicMock = mock_open(read_data="".join(MockCurva))
    with patch("builtins.open", m_leitura):
        cf1 = Curva.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Curva.le_arquivo("")
        assert cf1 == cf2



# Rotinas de testes associadas ao arquivo penalid.dat do NEWAVE

from inewave.newave.penalid import Penalid


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.penalid import (
    MockPenalid,
)


def test_atributos_encontrados_penalid():
    m: MagicMock = mock_open(read_data="".join(MockPenalid))
    with patch("builtins.open", m):
        ad = Penalid.le_arquivo("")
        assert ad.penalidades is not None


def test_atributos_nao_encontrados_penalid():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Penalid.le_arquivo("")
        assert ad.penalidades is None


def test_eq_penalid():
    m: MagicMock = mock_open(read_data="".join(MockPenalid))
    with patch("builtins.open", m):
        cf1 = Penalid.le_arquivo("")
        cf2 = Penalid.le_arquivo("")
        assert cf1 == cf2


def test_neq_penalid():
    m: MagicMock = mock_open(read_data="".join(MockPenalid))
    with patch("builtins.open", m):
        cf1 = Penalid.le_arquivo("")
        cf2 = Penalid.le_arquivo("")
        cf2.penalidades.loc[0, 0] = 0
        assert cf1 != cf2


def test_leitura_escrita_penalid():
    m_leitura: MagicMock = mock_open(read_data="".join(MockPenalid))
    with patch("builtins.open", m_leitura):
        cf1 = Penalid.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Penalid.le_arquivo("")
        assert cf1 == cf2



MockRegistroEolicaGeracaoPatamar = "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/01; 2021/01; 2 ; 1.0496\n"

MockEolicaGeracao = [
    "& EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; CodEolica;PerIni ;PerFin ;Pat   ;PROFUNDIDADE\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/01; 2021/01; 1 ; 0.8800\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/01; 2021/01; 2 ; 1.0496\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/01; 2021/01; 3 ; 1.0421\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/02; 2021/02; 1 ; 0.8800\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/02; 2021/02; 2 ; 1.0496\n",
    "EOLICA-GERACAO-PROFUNDIDADE-PERIODO-PATAMAR; 1 ; 2021/02; 2021/02; 3 ; 1.0421\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t \n",
]



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from typing import List, IO
import pandas as pd  # type: ignore


class BlocoConfUTE(Section):
    """
    Bloco de informações das usinas cadastradas
    no arquivo do NEWAVE `conft.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ute = Line(
            [
                IntegerField(4, 1),
                LiteralField(12, 6),
                IntegerField(4, 21),
                LiteralField(2, 30),
                IntegerField(4, 35),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConfUTE):
            return False
        bloco: BlocoConfUTE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_utes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_utes, 0)
            col_nome = extrai_coluna_de_listas(dados_utes, 1)
            col_subsis = extrai_coluna_de_listas(dados_utes, 2)
            col_exis = extrai_coluna_de_listas(dados_utes, 3)
            col_clas = extrai_coluna_de_listas(dados_utes, 4)
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Subsistema": col_subsis,
                "Usina Existente": col_exis,
                "Classe": col_clas,
            }
            return pd.DataFrame(data=dados)

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Para cada usina, lê e processa as informações
        dados_utes: List[list] = []
        while True:
            linha = file.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                if len(dados_utes) > 0:
                    self.data = transforma_utes_em_tabela()
                break
            dados_ute = self.__linha_ute.read(linha)
            dados_utes.append(dados_ute)

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do conft.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(self.__linha_ute.write(lin.tolist()))



# Imports do próprio módulo

from inewave.config import (
    MAX_ANOS_HISTORICO,
    MAX_CONFIGURACOES,
    MAX_UEES,
    MESES_ABREV,
    MAX_REES,
    MESES_DF,
)

# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from datetime import date
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoSerieVentosUEE(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo às séries de ventos por usina.
    """

    BEGIN_PATTERN = "SERIE  DE VENTOS   DA USINA"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_uee = LiteralField(14, 57)
        self.__campo_cfg = IntegerField(5, 88)
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campo_energia: List[Field] = [
            FloatField(10, 4 + 11 * i, 2) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campo_energia)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieVentosUEE):
            return False
        bloco: BlocoSerieVentosUEE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["Ano"] + MESES_DF)
            df["UEE"] = uee
            df["Configuração"] = cfg
            df = df[["UEE", "Configuração", "Ano"] + MESES_DF]
            df = df.astype({"Ano": "int64"})
            return df

        # Identifica a usina e a configuração
        linha = file.readline()
        uee = self.__campo_uee.read(linha)
        cfg = self.__campo_cfg.read(linha)

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF) + 1))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)
            i += 1


class BlocoCorrelVentosUEE(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo às séries de autocorrelações
    dos ventos por usina.
    """

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE VENTOS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelVentosUEE):
            return False
        bloco: BlocoCorrelVentosUEE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            cols = [f"Lag {i}" for i in range(1, 12)]
            df = pd.DataFrame(tabela, columns=cols)
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            df["Data"] = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = df[["Data"] + cols]
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoSerieRuidosUEE(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo às séries de ruídos por usina.
    """

    BEGIN_PATTERN = "SERIE DE RUIDOS  - ANO:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_ano = LiteralField(5, 81)
        campos: List[Field] = [
            FloatField(11, 3 + 11 * i, 3, format="E")
            for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campos)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieRuidosUEE):
            return False
        bloco: BlocoSerieRuidosUEE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF)
            df["Ano"] = ano
            df["Série"] = list(range(i))
            df = df[["Ano", "Série"] + MESES_DF]
            return df

        # Identifica o ano em questão
        linha = file.readline()
        ano = self.__campo_ano.read(linha)

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)
            i += 1


class BlocoCorrelRuidosUEE(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo às séries de autocorrelações
    dos ruídos por usina.
    """

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE RUIDOS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelRuidosUEE):
            return False
        bloco: BlocoCorrelRuidosUEE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = sorted(anos_estudo)[0]
            # Descobre o último ano de estudo
            ultimo_ano_estudo = sorted(anos_estudo)[-1]
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            cols = [f"Lag {i}" for i in range(1, 12)]
            df = pd.DataFrame(tabela, columns=cols)
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            df["Data"] = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = df[["Data"] + cols]
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoCorrelEspacialAnualConfig(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo à correlação espacial histórica espacial
    anual por configuração.
    """

    BEGIN_PATTERN = "CORRELACAO ESPACIAL HISTORICA ANUAL"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_cfg = IntegerField(4, 68)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEspacialAnualConfig):
            return False
        bloco: BlocoCorrelEspacialAnualConfig = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=rees_uees)
            df["UEE"] = uees
            df["Configuração"] = cfg
            df = df[["Configuração", "UEE"] + rees_uees]
            return df

        # Identifica a configuração
        linha = file.readline()
        cfg = self.__campo_cfg.read(linha)

        file.readline()
        linha = file.readline()
        rees_uees = [r.strip() for r in linha.split(" ") if len(r) > 2]
        campo_uee: List[Field] = [LiteralField(12, 0)]
        campo_correl: List[Field] = [
            FloatField(13, 12 + 13 * i, 4) for i in range(len(rees_uees))
        ]
        self.__linha = Line(campo_uee + campo_correl)

        # Variáveis auxiliares
        uees: List[str] = []
        tabela = np.zeros((MAX_UEES, len(rees_uees)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            tabela[i, :] = dados[1:]
            uees.append(dados[0])
            i += 1


class BlocoCorrelEspacialMensalConfig(Block):
    """
    Bloco de informações do arquivo `parpeol.dat`
    relativo à correlação espacial histórica espacial
    mensal por configuração.
    """

    BEGIN_PATTERN = "CORRELACAO ESPACIAL HISTORICA MENSAL"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_cfg = IntegerField(4, 69)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEspacialMensalConfig):
            return False
        bloco: BlocoCorrelEspacialMensalConfig = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=uees_rees)
            df["UEE"] = uees
            df["Configuração"] = cfg
            df = df[["Configuração", "UEE"] + uees_rees]
            df = df.astype({"MES": "int64"})
            return df

        # Identifica a configuração
        linha = file.readline()
        cfg = self.__campo_cfg.read(linha)

        file.readline()
        linha = file.readline()
        uees_rees = [r.strip() for r in linha.split(" ") if len(r) > 2]
        campo_uee: List[Field] = [LiteralField(12, 0)]
        campo_mes: List[Field] = [IntegerField(2, 15)]
        campo_correl: List[Field] = [
            FloatField(8, 17 + 13 * i, 4) for i in range(len(uees_rees) - 1)
        ]
        self.__linha = Line(campo_uee + campo_mes + campo_correl)

        # Variáveis auxiliares
        tabela = np.zeros(
            (
                MAX_CONFIGURACOES * (MAX_UEES + MAX_REES) * len(MESES_DF),
                len(uees_rees),
            )
        )
        uees: List[str] = []
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                linha = file.readline()
                if len(linha) < 4:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                    break
                else:
                    continue
            # Senão, processa os dados
            dados_linha = self.__linha.read(linha)
            uees.append(dados_linha[0])
            tabela[i, :] = dados_linha[1:]
            i += 1



from inewave.newave.modelos.pmo import (
    BlocoConvergenciaPMO,
    BlocoEafPastTendenciaHidrolPMO,
    BlocoEafPastCfugaMedioPMO,
    BlocoConfiguracoesExpansaoPMO,
    BlocoMARSPMO,
    BlocoRiscoDeficitENSPMO,
    BlocoCustoOperacaoPMO,
    BlocoCustoOperacaoTotalPMO,
)

from inewave.newave import PMO


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.pmo import MockBlocoEafPastTendenciaHidrolPMO
from tests.mocks.arquivos.pmo import MockBlocoEafPastCfugaMedioPMO
from tests.mocks.arquivos.pmo import MockBlocoConvergenciaPMO
from tests.mocks.arquivos.pmo import (
    MockBlocoConfiguracoesExpansaoEntradaReservatorioPMO,
)
from tests.mocks.arquivos.pmo import MockBlocoMARSPMOInicial
from tests.mocks.arquivos.pmo import MockBlocoMARSPMOFinal
from tests.mocks.arquivos.pmo import MockBlocoRiscoDeficitENSPMO
from tests.mocks.arquivos.pmo import MockBlocoCustoOperacaoPMO
from tests.mocks.arquivos.pmo import MockBlocoCustoOperacaoTotalPMO
from tests.mocks.arquivos.pmo import MockPMO


def test_eafpast_tendencia_hidrologica():

    m: MagicMock = mock_open(
        read_data="".join(MockBlocoEafPastTendenciaHidrolPMO)
    )
    b = BlocoEafPastTendenciaHidrolPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 13
    assert b.data.iloc[0, 1] == 7196.96
    assert b.data.iloc[-1, -1] == 409.13


def test_eafpast_cfuga_medio():

    m: MagicMock = mock_open(read_data="".join(MockBlocoEafPastCfugaMedioPMO))
    b = BlocoEafPastCfugaMedioPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 13
    assert b.data.iloc[0, 1] == 7196.96
    assert b.data.iloc[-1, -1] == 409.13


def test_convergencia():

    m: MagicMock = mock_open(read_data="".join(MockBlocoConvergenciaPMO))
    b = BlocoConvergenciaPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert len(list(b.data.index)) == 135


def test_leitura_configs():
    m: MagicMock = mock_open(
        read_data="".join(MockBlocoConfiguracoesExpansaoEntradaReservatorioPMO)
    )
    b = BlocoConfiguracoesExpansaoPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 0] == 2021
    assert b.data.iloc[-1, 0] == 2030
    assert b.data.iloc[0, 1] == 1
    assert b.data.iloc[-1, -1] == 60


def test_leitura_retas_perdas_inicial():
    m: MagicMock = mock_open(read_data="".join(MockBlocoMARSPMOInicial))
    b = BlocoMARSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 2] == 0.074069


def test_leitura_retas_perdas_final():
    m: MagicMock = mock_open(read_data="".join(MockBlocoMARSPMOFinal))
    b = BlocoMARSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 2] == 0.074069


def test_leitura_ens_deficit():
    m: MagicMock = mock_open(read_data="".join(MockBlocoRiscoDeficitENSPMO))
    b = BlocoRiscoDeficitENSPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert len(list(b.data.index)) == 5
    assert b.data.iloc[4, 1] == 0.0
    assert b.data.iloc[4, 2] == 0.0
    assert b.data.iloc[4, 3] == 0.0
    assert b.data.iloc[4, 4] == 0.0


def test_leitura_tabelas_custos():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoOperacaoPMO))
    b = BlocoCustoOperacaoPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.iloc[0, 1] == 21887.91


def test_leitura_custo_total():
    m: MagicMock = mock_open(read_data="".join(MockBlocoCustoOperacaoTotalPMO))
    b = BlocoCustoOperacaoTotalPMO()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data[0] == 22323.17
    assert b.data[1] == 107.00


def test_atributos_encontrados_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo = PMO.le_arquivo("")
        assert pmo.eafpast_tendencia_hidrologica is not None
        assert pmo.eafpast_cfuga_medio is not None
        assert pmo.convergencia is not None
        assert pmo.configuracoes_alteracao_potencia is not None
        assert pmo.configuracoes_entrada_reservatorio is not None
        assert pmo.configuracoes_qualquer_modificacao is not None
        assert pmo.retas_perdas_engolimento(1) is not None
        assert pmo.retas_perdas_engolimento(2) is not None
        assert pmo.risco_deficit_ens is not None
        assert pmo.custo_operacao_series_simuladas is not None
        assert pmo.custo_operacao_referenciado_primeiro_mes is not None
        assert pmo.valor_esperado_periodo_estudo is not None
        assert pmo.custo_operacao_total is not None
        assert pmo.desvio_custo_operacao_total is not None


def test_atributos_nao_encontrados_pmo():
    m: MagicMock = mock_open(read_data="".join(MockBlocoConvergenciaPMO))
    with patch("builtins.open", m):
        pmo = PMO.le_arquivo("")
        assert pmo.convergencia is not None
        assert pmo.custo_operacao_series_simuladas is None
        assert pmo.custo_operacao_total is None
        assert pmo.desvio_custo_operacao_total is None


def test_eq_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo1 = PMO.le_arquivo("")
        pmo2 = PMO.le_arquivo("")
        assert pmo1 == pmo2


def test_neq_pmo():
    m: MagicMock = mock_open(read_data="".join(MockPMO))
    with patch("builtins.open", m):
        pmo1 = PMO.le_arquivo("")
        pmo2 = PMO.le_arquivo("")
        pmo2.configuracoes_alteracao_potencia
        assert pmo1 == pmo2



MockRegistroEolicaHistoricoHorizonte = (
    "EOLICA-HISTORICO-VENTO-HORIZONTE ; 1979/01 ; 2016/01\n"
)

MockRegistroEolicaHistorico = (
    "EOLICA-HISTORICO-VENTO; 1; 1979/01; 1979/02;    3.43; 1.0\n"
)

MockEolicaHistorico = [
    "& EOLICA-HISTORICO-VENTO-HORIZONTE ; DataInicial ; DataFinal\n",
    "EOLICA-HISTORICO-VENTO-HORIZONTE ; 1979/01 ; 2016/01\n",
    "\n",
    "\n",
    "& EOLICA-HISTORICO-VENTO ; CodigoEolica ; DataInicial ; DataFinal ; MagnitudeVelocidadeVento ; DirecaoVento \n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/01; 1979/02;    3.43; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/02; 1979/03;    4.28; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/03; 1979/04;    4.67; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/04; 1979/05;    4.50; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/05; 1979/06;    4.77; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/06; 1979/07;    6.09; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/07; 1979/08;    5.66; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/08; 1979/09;    6.21; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/09; 1979/10;    6.75; 1.0\n",
    "EOLICA-HISTORICO-VENTO; 1; 1979/10; 1979/11;    5.64; 1.0\n",
]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.varmpuh import VarmAnos


class VarmpUH(ArquivoUsina):
    """
    Armazena os dados das saídas referentes aos armazenamentos em
    percentual por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `varmpuh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        VarmAnos,
    ]

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="varmpuh001.out"
    ) -> "VarmpUH":
        return cls.read(diretorio, nome_arquivo)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="varmpuh001.out"):
        self.write(diretorio, nome_arquivo)



from inewave.newave.modelos.arquivos import BlocoNomesArquivos
from inewave.newave import Arquivos

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.arquivos import MockArquivos


def test_bloco_nomes_arquivos():

    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    b = BlocoNomesArquivos()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 41
    assert b.data.shape[1] == 2
    assert b.data.iloc[0, 1] == "dger.dat"
    assert b.data.iloc[-1, 1] == "clasgas.dat"


def test_atributos_encontrados_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad = Arquivos.le_arquivo("")
        assert ad.dger is not None
        assert ad.sistema is not None
        assert ad.confhd is not None
        assert ad.modif is not None
        assert ad.conft is not None
        assert ad.term is not None
        assert ad.clast is not None
        assert ad.exph is not None
        assert ad.expt is not None
        assert ad.patamar is not None
        assert ad.cortes is not None
        assert ad.cortesh is not None
        assert ad.pmo is not None
        assert ad.parp is not None
        assert ad.forward is not None
        assert ad.forwardh is not None
        assert ad.shist is not None
        assert ad.manutt is not None
        assert ad.newdesp is not None
        assert ad.vazpast is not None
        assert ad.itaipu is not None
        assert ad.bid is not None
        assert ad.c_adic is not None
        assert ad.perda is not None
        assert ad.gtminpat is not None
        assert ad.elnino is not None
        assert ad.ensoaux is not None
        assert ad.dsvagua is not None
        assert ad.penalid is not None
        assert ad.curva is not None
        assert ad.agrint is not None
        assert ad.adterm is not None
        assert ad.ghmin is not None
        assert ad.sar is not None
        assert ad.cvar is not None
        assert ad.ree is not None
        assert ad.re is not None
        assert ad.tecno is not None
        assert ad.abertura is not None
        assert ad.gee is not None
        assert ad.clasgas is not None


def test_atributos_nao_encontrados_arquivos():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Arquivos.le_arquivo("")
        assert ad.dger is None
        assert ad.sistema is None
        assert ad.confhd is None
        assert ad.modif is None
        assert ad.conft is None
        assert ad.term is None
        assert ad.clast is None
        assert ad.exph is None
        assert ad.expt is None
        assert ad.patamar is None
        assert ad.cortes is None
        assert ad.cortesh is None
        assert ad.pmo is None
        assert ad.parp is None
        assert ad.forward is None
        assert ad.forwardh is None
        assert ad.shist is None
        assert ad.manutt is None
        assert ad.newdesp is None
        assert ad.vazpast is None
        assert ad.itaipu is None
        assert ad.bid is None
        assert ad.c_adic is None
        assert ad.perda is None
        assert ad.gtminpat is None
        assert ad.elnino is None
        assert ad.ensoaux is None
        assert ad.dsvagua is None
        assert ad.penalid is None
        assert ad.curva is None
        assert ad.agrint is None
        assert ad.adterm is None
        assert ad.ghmin is None
        assert ad.sar is None
        assert ad.cvar is None
        assert ad.ree is None
        assert ad.re is None
        assert ad.tecno is None
        assert ad.abertura is None
        assert ad.gee is None
        assert ad.clasgas is None


def test_eq_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad1 = Arquivos.le_arquivo("")
        ad2 = Arquivos.le_arquivo("")
        assert ad1 == ad2


def test_neq_arquivos():
    m: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m):
        ad1 = Arquivos.le_arquivo("")
        ad2 = Arquivos.le_arquivo("")
        ad2.dger = "teste"
        assert ad1 != ad2


def test_leitura_escrita_arquivos():
    m_leitura: MagicMock = mock_open(read_data="".join(MockArquivos))
    with patch("builtins.open", m_leitura):
        ad1 = Arquivos.le_arquivo("")
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        ad1.escreve_arquivo("", "")
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(2, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        ad2 = Arquivos.le_arquivo("")
        assert ad1 == ad2



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from typing import IO, List
import pandas as pd  # type: ignore


class BlocoNomesArquivos(Section):
    """
    Bloco de informações do arquivo de
    entrada do NEWAVE `arquivos.dat`.
    """

    FIM_BLOCO = " 9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line([LiteralField(30, 0), LiteralField(40, 30)])

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoNomesArquivos):
            return False
        bloco: BlocoNomesArquivos = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(data={"Legenda": legendas, "Nome": nomes})
            return df

        legendas: List[str] = []
        nomes: List[str] = []
        while True:
            linha = file.readline()
            if len(linha) < 3:
                self.data = converte_tabela_em_df()
                break
            dados = self.__linha.read(linha)
            legendas.append(dados[0])
            nomes.append(dados[1])

    # Override
    def write(self, file: IO):
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do arquivos.dat não foram lidos")
        for _, linha in self.data.iterrows():
            file.write(self.__linha.write(linha.tolist()))



# Inclui os membros

from .abertura import Abertura  # noqa
from .adterm import AdTerm  # noqa
from .agrint import AgrInt  # noqa
from .arquivos import Arquivos  # noqa
from .bid import BID  # noqa
from .cadic import CAdic  # noqa
from .caso import Caso  # noqa
from .clasgas import ClasGas  # noqa
from .clast import ClasT  # noqa
from .confhd import Confhd  # noqa
from .conft import ConfT  # noqa
from .curva import Curva  # noqa
from .cvar import CVAR  # noqa
from .dger import DGer  # noqa
from .dsvagua import DSVAgua  # noqa
from .eafpast import EafPast  # noqa
from .elnino import ElNino  # noqa
from .ensoaux import ENSOAux  # noqa
from .exph import Exph  # noqa
from .expt import Expt  # noqa
from .gee import GEE  # noqa
from .ghmin import GHMin  # noqa
from .gtminpat import GTMinPat  # noqa
from .hidr import Hidr  # noqa
from .itaipu import Itaipu  # noqa
from .manutt import ManutT  # noqa
from .modif import Modif  # noqa
from .newavetim import NewaveTim  # noqa
from .parp import PARp  # noqa
from .parpeol import PARpeol  # noqa
from .patamar import Patamar  # noqa
from .penalid import Penalid  # noqa
from .perda import Perda  # noqa
from .pmo import PMO  # noqa
from .re import RE  # noqa
from .ree import REE  # noqa
from .sar import SAR  # noqa
from .shist import Shist  # noqa
from .sistema import Sistema  # noqa
from .tecno import Tecno  # noqa
from .term import Term  # noqa
from .vazoes import Vazoes  # noqa
from .vazpast import VazPast  # noqa
from .eolicacadastro import EolicaCadastro  # noqa
from .eolicaconfiguracao import EolicaConfiguracao  # noqa
from .eolicafte import EolicaFTE  # noqa
from .eolicageracao import EolicaGeracao  # noqa
from .eolicahistorico import EolicaHistorico  # noqa
from .eolicasubmercado import EolicaSubmercado  # noqa



from inewave.config import MAX_ANOS_ESTUDO, MAX_SUBMERCADOS, MESES_DF

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.literalfield import LiteralField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoConfiguracoesPenalizacaoCurva(Section):
    """
    Bloco de informações das usinas cadastradas
    no arquivo do NEWAVE `conft.dat`.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                IntegerField(3, 5),
                IntegerField(3, 9),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoConfiguracoesPenalizacaoCurva):
            return False
        bloco: BlocoConfiguracoesPenalizacaoCurva = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.__cabecalhos.append(file.readline())

        self.data = self.__linha.read(file.readline())

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, list):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")
        file.write(self.__linha.write(self.data))


class BlocoPenalidadesViolacaoREECurva(Section):
    """
    Bloco com informações das penalidades por violação para cada
    REE.
    """

    FIM_BLOCO = " 999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                IntegerField(3, 1),
                FloatField(7, 11, 2),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoPenalidadesViolacaoREECurva):
            return False
        bloco: BlocoPenalidadesViolacaoREECurva = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["Sistema", "Custo"])
            df = df.astype({"Sistema": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        i = 0
        tabela = np.zeros((MAX_SUBMERCADOS, 2))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                len(linha) < 3
                or BlocoPenalidadesViolacaoREECurva.FIM_BLOCO in linha
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            tabela[i, :] = self.__linha.read(linha)
            i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        for _, lin in self.data.iterrows():
            file.write(
                self.__linha.write([int(lin["Sistema"])] + [lin["Custo"]])
            )
        file.write(BlocoPenalidadesViolacaoREECurva.FIM_BLOCO + "\n")


class BlocoCurvaSegurancaSubsistema(Section):
    """
    Bloco com informações da curva de segurança de operação por mês/ano
    e por subsistema.
    """

    FIM_BLOCO = "9999"

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_subsis = Line([IntegerField(3, 1)])
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campos_curva: List[Field] = [
            FloatField(5, i * 6 + 6, 1) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campos_curva)
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCurvaSegurancaSubsistema):
            return False
        bloco: BlocoCurvaSegurancaSubsistema = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=["REE", "Ano"] + MESES_DF)
            df = df.astype({"REE": "int64", "Ano": "int64"})
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            self.__cabecalhos.append(file.readline())

        i = 0
        subsis_atual = 0
        tabela = np.zeros(
            (MAX_SUBMERCADOS * MAX_ANOS_ESTUDO, len(MESES_DF) + 2)
        )
        while True:
            linha = file.readline()
            # Confere se terminaram
            if (
                len(linha) < 3
                or BlocoCurvaSegurancaSubsistema.FIM_BLOCO in linha
            ):
                # Converte para df e salva na variável
                if i > 0:
                    tabela = tabela[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            if len(linha) < 10:
                subsis_atual = self.__linha_subsis.read(linha)[0]
            else:
                tabela[i, 0] = subsis_atual
                tabela[i, 1:] = self.__linha.read(linha)
                i += 1

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        ultimo_ree = 0
        for _, linha in self.data.iterrows():
            linha_lida: pd.Series = linha
            if linha_lida["REE"] != ultimo_ree:
                ultimo_ree = linha_lida["REE"]
                file.write(self.__linha_subsis.write([int(ultimo_ree)]))
            file.write(
                self.__linha.write(
                    [int(linha_lida["Ano"])] + linha_lida[MESES_DF].tolist()
                )
            )

        file.write(BlocoCurvaSegurancaSubsistema.FIM_BLOCO + "\n")


class BlocoMaximoIteracoesProcessoIterativoEtapa2(Section):
    """
    Bloco com informações do número máximo de iterações da
    segunda etapa do processo iterativo no cálculo da curva de aversão.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [LiteralField(26, 0), IntegerField(6, 28), LiteralField(46, 39)]
        )
        self.__cabecalhos: List[str] = []
        self.__campo: str = ""
        self.__comentario: str = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoMaximoIteracoesProcessoIterativoEtapa2):
            return False
        bloco: BlocoMaximoIteracoesProcessoIterativoEtapa2 = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):

        # Salta as linhas adicionais
        self.__cabecalhos.append(file.readline())

        self.__campo, self.data, self.__comentario = self.__linha.read(
            file.readline()
        )

    # Override
    def write(self, file: IO):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, int):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        file.write(
            self.__linha.write([self.__campo, self.data, self.__comentario])
        )


class BlocoIteracaoAPartirProcessoIterativoEtapa2(Section):
    """
    Bloco com informações da iteração a partir da qual ocorre a
    segunda etapa do processo iterativo no cálculo da curva de aversão.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [LiteralField(26, 0), IntegerField(6, 28), LiteralField(46, 39)]
        )
        self.__campo: str = ""
        self.__comentario: str = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoIteracaoAPartirProcessoIterativoEtapa2):
            return False
        bloco: BlocoIteracaoAPartirProcessoIterativoEtapa2 = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):

        self.__campo, self.data, self.__comentario = self.__linha.read(
            file.readline()
        )

    # Override
    def write(self, file: IO):
        if not isinstance(self.data, int):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        file.write(
            self.__linha.write([self.__campo, self.data, self.__comentario])
        )


class BlocoToleranciaProcessoIterativoEtapa2(Section):
    """
    Bloco com informações da tolerância para a
    segunda etapa do processo iterativo no cálculo da curva de aversão.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [LiteralField(26, 0), FloatField(6, 28, 3), LiteralField(46, 39)]
        )
        self.__campo: str = ""
        self.__comentario: str = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoToleranciaProcessoIterativoEtapa2):
            return False
        bloco: BlocoToleranciaProcessoIterativoEtapa2 = o
        if not all(
            [
                isinstance(self.data, float),
                isinstance(o.data, float),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):

        self.__campo, self.data, self.__comentario = self.__linha.read(
            file.readline()
        )

    # Override
    def write(self, file: IO):
        if not isinstance(self.data, float):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        file.write(
            self.__linha.write([self.__campo, self.data, self.__comentario])
        )


class BlocoImpressaoRelatorioProcessoIterativoEtapa2(Section):
    """
    Bloco com informações da impressão de relatório na
    segunda etapa do processo iterativo no cálculo da curva de aversão.
    """

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [LiteralField(26, 0), IntegerField(6, 28), LiteralField(46, 39)]
        )
        self.__campo: str = ""
        self.__comentario: str = ""

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoImpressaoRelatorioProcessoIterativoEtapa2):
            return False
        bloco: BlocoImpressaoRelatorioProcessoIterativoEtapa2 = o
        if not all(
            [
                isinstance(self.data, int),
                isinstance(o.data, int),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO):
        self.__campo, self.data, self.__comentario = self.__linha.read(
            file.readline()
        )

    # Override
    def write(self, file: IO):
        if not isinstance(self.data, int):
            raise ValueError("Dados do curva.dat não foram lidos com sucesso")

        file.write(
            self.__linha.write([self.__campo, self.data, self.__comentario])
        )



MockDger = [
    "PMO JANEIRO - 2021  22/12/2020 12:43:55  Niveis para 26/12 NW Versao 27.5_CPAMP\n",
    "TIPO DE EXECUCAO        1   (1:EXECUCAO COMPLETA; 0:SIMULACAO FINAL)\n",
    "DURACAO DO PERIODO      1\n",
    "No. DE ANOS DO EST      5\n",
    "MES INICIO PRE-EST      1\n",
    "MES INICIO DO ESTUDO    1\n",
    "ANO INICIO DO ESTUDO 2021\n",
    "No. DE ANOS PRE         0\n",
    "No. DE ANOS POS         5\n",
    "No. DE ANOS POS FINAL   0\n",
    "IMPRIME DADOS           1\n",
    "IMPRIME MERCADOS        1\n",
    "IMPRIME ENERGIAS        1\n",
    "IMPRIME M. ESTOCAS      1\n",
    "IMPRIME SUBSISTEMA      1\n",
    "No MAX. DE ITER.       45\n",
    "No DE SIM. FORWARD    200\n",
    "No DE ABERTURAS        20\n",
    "No DE SERIES SINT.   2000\n",
    "ORDEM MAX. PAR(P)       6\n",
    "ANO INICIAL HIST.    1931   0\n",
    "CALCULA VOL.INICIAL     1   0=USA REG 20 ; 1= CALCULA EARM. INICIAL\n",
    "VOLUME INICIAL  -%   XXX.X  XXX.X  XXX.X  XXX.X  XXX.X  \n",
    " POR SUBSISTEMA        0.0    0.0    0.0    0.0    0.0  \n",
    "TOLERANCIA      -%    95.0\n",
    "TAXA DE DESCONTO-%    12.0\n",
    "TIPO SIMUL. FINAL       1   1  (=0 NAO SIMULA; =1 S.SINT.; =2 S.HIST.; =3 CONSIST/ =0 SIM.REE; 1=SIM.INDIV)\n",
    "IMPRESSAO DA OPER       1   (=0 SINOPSE; =1 OP. DETALHADA)\n",
    "IMPRESSAO DA CONVERG.   1   (=0 CONVERGENCIA FINAL APENAS, =1 TOTAL)\n",
    "INTERVALO P/ GRAVAR     1   SERIES SIMULADAS ( 40  SERIES GRAVADAS )\n",
    "No. MIN. ITER.         30\n",
    "RACIONAMENTO PREVENT.   0   (=0 NAO CONSIDERA NA SIMULACAO FINAL; 1=CONSIDERA)\n",
    "No. ANOS MANUT.UTE'S    1   (=0 NAO CONSIDERA, =1 ANO, =2 ANOS)\n",
    "TENDENCIA HIDROLOGICA   2    2   (1� FCF / 2� SF =0 NAO CONDIC., =1 P/ SUBSISTEMA, =2 P/ POSTO)\n",
    "RESTRICA0 DE ITAIPU     0   (=0 NAO CONSIDERA, =1 CONSIDERA)\n",
    "BID                     0   (=0 NAO CONSIDERA, =1 CONSIDERA)\n",
    "PERDAS P/ TRANSMISSAO   0   (=0 NAO CONSIDERA, =1 CONSIDERA)\n",
    "EL NINO                 0   (=0 NAO CONSIDERA, =1 CONSIDERA)\n",
    "ENSO INDEX              0   (FUNCAO NAO IMPLEMENTADA)\n",
    "DURACAO POR PATAMAR     1   (=0 SAZONAL, =1 VARIAVEL POR ANO)\n",
    "OUTROS USOS DA AGUA     1   (=0 NAO CONSIDERA, =1 CONSIDERA)\n",
    "CORRECAO DESVIO         1   (=0 CONSTANTE; =1 VARIAVEL COM O ARMAZENAMENTO)\n",
    "C.AVERSAO/PENAL.VMINP   1   (=0 SEM CAR E VMINP; =1  CAR E/OU VMINP)\n",
    "TIPO DE GERACAO ENAS    0   (=0 RUIDOS FW SORTEADOS DA BW E COMPENSACAO CORREL.ESPACIAL; =1 COMPENS.BW; =2 COMPENS.BW E FW)\n",
    "RISCO DE DEFICIT        1.  2.5\n",
    "ITERACAO P/SIM.FINAL    0   (=0 CONSIDERA TODAS AS ITERACOES)\n",
    "AGRUPAMENTO LIVRE       1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "EQUALIZACAO PEN.INT.    1   (FLAG DESABILITADO)\n",
    "REPRESENT.SUBMOT.       2   (=0 NAO CONSIDERA , =1 SUBSISTEMA, 2=USINA)\n",
    "ORDENACAO AUTOMATICA    0   (=0 NAO CONSIDERA; =1 CONSIDERA)\n",
    "CONS. CARGA ADICIONAL   1   (=0 NAO CONSIDERA; =1 CONSIDERA)\n",
    "DELTA ZSUP             10   (VALOR EM PERCENTUAL)\n",
    "DELTA ZINF            0.2   (VALOR EM PERCENTUAL)\n",
    "DELTAS CONSECUT.        3\n",
    "DESP. ANTEC.  GNL       1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "MODIF.AUTOM.ADTERM      1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "CONSIDERA GHMIN         1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "S.F. COM DATA           0\n",
    "GER.PLs E NV1 E NV2     0    0    0    0    0    (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "SAR                     0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "CVAR                    1   (=0 NAO CONSIDERA , =1 CONSIDERA CTE TEMPO, =2 CONSIDERA VARIAVEL NO TEMP)\n",
    "CONS. ZSUP MIN. CONV.   0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "DESCONSIDERA VAZMIN     0   (=0 NAO , =1 SIM)\n",
    "RESTRICOES ELETRICAS    1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "SELECAO DE CORTES       1    1    (=0 NAO CONSIDERA NA ETAPA BACKWARD, =1 CONSIDERA NA ETAPA BACKWARD; =0 NAO CONSIDERA NA ETAPA FORWARD, =1 CONSIDERA NA ETAPA FORWARD)\n",
    "JANELA DE CORTES        0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "REAMOST. CENARIOS    0001 0001 0001    (UTILIZA REAMOSTRAGEM: =0 NAO; =1 SIM     TIPO: =0 RECOMB; =1 PLENA     PASSO: 0 - 45 )\n",
    "CONVERGE NO ZERO        0   (=0 CONVERGENCIA TRADICIONAL, =1 CONVERGENCIA CALCULADA NO ZERO)\n",
    "CONSULTA FCF            0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "IMPRESSAO ENA        0001   (=0 NAO IMPRIME , =1 IMPRIME)\n",
    "IMP. CATIVO S.FINAL  0000   (=0 NAO IMPRIME , =1 IMPRIME)\n",
    "REP. AGREGACAO       0001   (=0 MAIS PROXIMO, =1 CENTROIDE)\n",
    "MATRIZ CORR.ESPACIAL 0001   (=0 ANUAL, =1 MENSAL)\n",
    "DESCONS. CONV. ESTAT 0001   (=0 NAO, =1 SIM)\n",
    "MOMENTO REAMOSTRAGEM 0001   (=0 BACKWARD, =1 FORWARD)\n",
    "ARQUIVOS ENA         0000   (=0 APAGA APOS EXECUCAO, =1 MANTEM)\n",
    "INICIO TESTE CONVERG.   1   (=0 PRIMEIRA ITERACAO, =1 ITERACAO MINIMA)\n",
    "SAZ. VMINT PER. EST.    0   (=0 PRE E POS NAO SAZONAIS, =1 PRE E POS SAZONAIS)\n",
    "SAZ. VMAXT PER. EST.    0   (=0 PRE E POS NAO SAZONAIS, =1 PRE E POS SAZONAIS)\n",
    "SAZ. VMINP PER. EST.    0   (=0 PRE E POS NAO SAZONAIS, =1 PRE E POS SAZONAIS)\n",
    "SAZ. CFUGA E CMONT      0   (=0 PRE E POS NAO SAZONAIS, =1 PRE E POS SAZONAIS)\n",
    "REST. EMISSAO GEE       0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "AFLUENCIA ANUAL PARP    3    0   (=0 NAO CONSIDERA , =1 CONSIDERA. PDDE SEM ABRIR X NA DERIVACAO DOS CORTES, =2 CONSIDERA. PDDE ABRINDO X NA DERIVACAO DOS CORTES COM APROX. DE 1/6 X(t-1) PARA ENA(t-12), =3 CONSIDERA. PDDE EXATA COM 12 EIXOS PARA AS AFLUENCIAS PASSADAS; REDUCAO DA ORDEM: =0 CONSIDERA,  =1 NAO CONSIDERA, =2 CONSIDERA COM IMPRESSAO RELATORIO)\n",
    "REST. FORNEC. GAS       0   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "MEM. CALCULO CORTES     0   (=0 NAO IMPRIME, =1 IMPRIME PARA ESCOLHA ESPECIFICADA NO ARQUIVO dbgcortes.dat)\n",
    "GERACAO EOLICA          1 0.0063       (=0 NAO CONSIDERA, =1 CONSIDERA; PENALIDADE DO CORTE DE GERACAO EOLICA)\n",
    "COMP. COR. CRUZ.        1   (=0 CORR. SOMENTE AFL.; =1 CORR.AFL. E VENTOS; =2 NAO CORRIGE)\n",
    "REST. TURBINAMENTO      1   (=0 NAO CONSIDERA , =1 CONSIDERA REST. DE TURB. MAXIMO E MINIMO, =2 CONSIDERA APENAS MAXIMO, =3 CONSIDERA APENAS MINIMO)\n",
    "REST. DEFL. MAXIMA      1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "BASE PLS BACKWARD       1   (=0 CALCULA, 1 = USA DA FORWARD ANTERIOR)\n",
    "ESTADOS GER. CORTES  0001   (=0 IMPRIME cortese.dat, =1 NAO IMPRIME cortese.dat)\n",
    "SEMENTE FORWARD      0000   (=0 A 1000, 0 = ORIGINAL)\n",
    "SEMENTE BACWARD      0000   (=0 A 1000, 0 = ORIGINAL)\n",
    "REST.LPP TURB.MAX REE   1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "REST.LPP DEFL.MAX REE   1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "REST.LPP TURB.MAX UHE   1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "REST.LPP DEFL.MAX UHE   1   (=0 NAO CONSIDERA , =1 CONSIDERA)\n",
    "",
]
